/*=========================================================================

  Program:   Visualization Toolkit
  Module:    $RCSfile: vtkSQ3SubdivisionFilter.cxx,v $
  Language:  C++
  Date:      $Date: 2002/12/17 02:05:38 $
  Version:   $Revision: 1.20 $

  Copyright (c) 1993-2002 Ken Martin, Will Schroeder, Bill Lorensen 
  All rights reserved.
  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even 
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/
#include "StdAfx.h"

#include "vtkSQ3SubdivisionFilter.h"

#include "vtkObjectFactory.h"
#include "vtkCellArray.h"
#include "vtkCellData.h"
#include "vtkEdgeTable.h"
#include "vtkPointData.h"
#include "vtkPolyData.h"

vtkCxxRevisionMacro(vtkSQ3SubdivisionFilter, "$Revision: 1.20 $");
vtkStandardNewMacro(vtkSQ3SubdivisionFilter);

// Construct object with number of subdivisions set to 1.
/*
vtkSQ3SubdivisionFilter::vtkSQ3SubdivisionFilter()
{
}
*/

void vtkSQ3SubdivisionFilter::Execute()
{
  vtkIdType numPts, numCells;
  int level;
  vtkPoints *outputPts;
  vtkCellArray *outputPolys;
  vtkPolyData *input = this->GetInput();
  vtkPolyData *output = this->GetOutput();
  vtkPointData *outputPD;
  vtkCellData *outputCD;

  vtkDebugMacro(<< "Generating subdivision surface using interpolating scheme");

  
  if (input == NULL)
    {
    vtkErrorMacro(<<"Input is NULL");
    return;
    }

  numPts=input->GetNumberOfPoints();
  numCells=input->GetNumberOfCells();

  if (numPts < 1 || numCells < 1)
    {
    vtkDebugMacro(<<"No data to interpolate!");
    return;
    }

  //
  // Initialize and check input
  //

  vtkPolyData *inputDS = vtkPolyData::New();
  inputDS->CopyStructure (input);
  inputDS->GetPointData()->PassData(input->GetPointData());
  inputDS->GetCellData()->PassData(input->GetCellData());

  for (level = 0; level < this->NumberOfSubdivisions; level++)
    {
    // Generate topology  for the input dataset
    inputDS->BuildLinks();
    numCells = inputDS->GetNumberOfCells ();

    // Copy points from input. The new points will include the old points
    // and points calculated by the subdivision algorithm
    outputPts = vtkPoints::New();
    //outputPts->GetData()->DeepCopy(inputDS->GetPoints()->GetData());
		outputPts->Allocate (numPts);

    // Copy pointdata structure from input
    outputPD = vtkPointData::New();
    outputPD->CopyAllocate(inputDS->GetPointData(),
                           2*inputDS->GetNumberOfPoints());

    // Copy celldata structure from input
    outputCD = vtkCellData::New();
    outputCD->CopyAllocate(inputDS->GetCellData(),3*numCells);
		//outputCD->Print(*os);

    // Create triangles
    outputPolys = vtkCellArray::New();
    outputPolys->Allocate(outputPolys->EstimateSize(4*numCells,3));

		// Do the split
    this->GenerateSubdivisionPoints (inputDS, outputPts, outputPD, outputPolys, outputCD);

    // start the next iteration with the input set to the output we just created
    inputDS->Delete();
    inputDS = vtkPolyData::New();
    inputDS->SetPoints(outputPts); outputPts->Delete();
    inputDS->SetPolys(outputPolys); outputPolys->Delete();
    inputDS->GetPointData()->PassData(outputPD); outputPD->Delete();
    inputDS->GetCellData()->PassData(outputCD); outputCD->Delete();
    inputDS->Squeeze();
    } // each level

  output->SetPoints(inputDS->GetPoints());
  output->SetPolys(inputDS->GetPolys());
  output->GetPointData()->PassData(inputDS->GetPointData());
  output->GetCellData()->PassData(inputDS->GetCellData());
  inputDS->Delete();
}

bool vtkSQ3SubdivisionFilter::generatePoint(vtkIdType cellId, vtkIdType &newId, vtkIdList *cellsVisited, 
																						vtkIdList *cellsPoints, vtkIdType *pts, vtkPolyData *inputDS, 
																						vtkPointData *outputPD, vtkPoints *outputPts)
{
	//vtkIdType newId;
  vtkPoints *inputPts=inputDS->GetPoints();
  vtkIdList *pointIds = vtkIdList::New();
  vtkPointData *inputPD=inputDS->GetPointData();
	static float weights[3] = {.33333, .33333, .33334};

	pointIds->SetNumberOfIds(3);
	newId = cellsPoints->GetId(cellId);

	if ( cellsVisited->GetId(cellId) == 0 ) {
		outputPD->CopyData(inputPD, pts[0], pts[0]);
		outputPD->CopyData(inputPD, pts[1], pts[1]);
		outputPD->CopyData(inputPD, pts[2], pts[2]);

		pointIds->SetId(0,pts[0]);
		pointIds->SetId(1,pts[1]);
		pointIds->SetId(2,pts[2]);
		newId = this->InterpolatePosition (inputPts, outputPts, pointIds, weights);
		outputPD->InterpolatePoint (inputPD, newId, pointIds, weights);

		cellsVisited->SetId(cellId, 1);
		cellsPoints->SetId(cellId, newId);

		return true;

		/*
		(*os) << "-------------------- Cell: " << cellId << "------------------\n";
		//(*os) << ">>>> Edge Data:\n";
		//edgeData->Print(*os);
		(*os) << ">>>> Edge elements:\n";
		for(int t=0; t<edgeData->GetNumberOfTuples(); ++t) {
			for(int c=0; c<edgeData->GetNumberOfComponents(); ++c)
				(*os) << "Tuple[" << t << "]::Comp[" << c << "] = " << edgeData->GetComponent(t, c) << endl;
		}
		(*os) << ">>>> End of edge elements\n";
		(*os) << ">>>> Points:\n";
		outputPts->Print(*os);
		for(int i=0; i<outputPts->GetNumberOfPoints(); ++i)
			(*os) << "outputPts[" << i << "]= (" 
						<< outputPts->GetPoint(i)[0] << "," 
						<< outputPts->GetPoint(i)[1] << "," 
						<< outputPts->GetPoint(i)[2] << ")\n";
		(*os) << ">>>> end of Points\n";

		(*os) << endl << endl;
		*/
	} else {
		return false;
	}
}

void vtkSQ3SubdivisionFilter::GenerateSubdivisionPoints (vtkPolyData *inputDS, 
																												 vtkPoints *outputPts, 
																												 vtkPointData *outputPD,
																												 vtkCellArray *outputPolys, 
																												 vtkCellData *outputCD)
{
  vtkIdType			*pts = 0;
  vtkIdType			npts, cellId, newId;
  vtkCellArray	*inputPolys=inputDS->GetPolys();
  vtkIdList			*cellIds = vtkIdList::New();
	vtkIdType			cell = 0;
	vtkIdList			*cellsVisited = vtkIdList::New();
	vtkIdList			*cellsPoints = vtkIdList::New();
	vtkIdType			newCellPts[3];
  vtkCellData		*inputCD = inputDS->GetCellData();
	vtkEdgeTable	*edgeTable = vtkEdgeTable::New();
	vtkIdType			cellCount = inputDS->GetNumberOfCells();
	vtkIdList			*stencil = vtkIdList::New();
	vtkPoints			*inputPts=inputDS->GetPoints();
  vtkPointData	*inputPD=inputDS->GetPointData();

	cellsVisited->SetNumberOfIds(cellCount);
	cellsPoints->SetNumberOfIds(cellCount);
	for(int i=0; i<cellCount; ++i) {
		cellsVisited->SetId(i, 0);
		cellsPoints->SetId(i, -1);
	}

	vtkIdType numET = inputDS->GetNumberOfPoints() + inputDS->GetNumberOfCells();
	edgeTable->InitEdgeInsertion(numET);

	vtkIdType numPts = inputDS->GetNumberOfPoints();
	float *weights = new float[256];
	for (vtkIdType ptId=0; ptId < numPts; ptId++)
	{
		this->GenerateEvenStencil (ptId, inputDS, stencil, weights);
    this->InterpolatePosition (inputPts, outputPts, stencil, weights);
		outputPD->InterpolatePoint (inputPD, ptId, stencil, weights);
	}

  // Generate new points for subdivisions surface
  for (cellId=0, inputPolys->InitTraversal();
       inputPolys->GetNextCell(npts, pts); cellId++)
  {
    if ( inputDS->GetCellType(cellId) != VTK_TRIANGLE )
    {
      continue;
    }

		bool actIsGen = generatePoint(cellId, newId, cellsVisited, cellsPoints, pts, inputDS, outputPD, outputPts);

		//bool haveMoreCells = true;
		vtkIdType p1 = pts[2], p2 = pts[0];
		for (int edgeId=0; edgeId < 3; edgeId++) {
			inputDS->GetCellEdgeNeighbors(cellId, p1, p2, cellIds);
			if ( cellIds->GetNumberOfIds() != 0) {
				vtkIdType *neighpts = 0;
				vtkIdType neighnpts;
				vtkIdType newnId;
				inputDS->GetCellPoints(cellIds->GetId(cell), neighnpts, neighpts);
				bool neighIsGen = generatePoint(cellIds->GetId(cell), newnId, cellsVisited, cellsPoints, neighpts, inputDS, outputPD, outputPts);

				if ( edgeTable->IsEdge(newId, newnId) == -1 ) {		// Todo: Do a real edge existence check :)
					vtkIdType newCellPts[3];
					int id;
					vtkIdType myNewId;

					edgeTable->InsertEdge(newId, newnId);

					id = 0;
					newCellPts[id++] = p1;
					newCellPts[id++] = newnId;
					newCellPts[id++] = newId;
					myNewId = outputPolys->InsertNextCell (3, newCellPts);
					outputCD->CopyData(inputCD, cellId, newId);
					id = 0;
					newCellPts[id++] = newId;
					newCellPts[id++] = newnId;
					newCellPts[id++] = p2;
					myNewId = outputPolys->InsertNextCell (3, newCellPts);
					outputCD->CopyData(inputCD, cellId, newId);
				}
			} else {
					vtkIdType newCellPts[3];
					int id;
					vtkIdType myNewId;

					id = 0;
					newCellPts[id++] = p1;
					newCellPts[id++] = p2;
					newCellPts[id++] = newId;
					myNewId = outputPolys->InsertNextCell (3, newCellPts);
					outputCD->CopyData(inputCD, cellId, myNewId);
			}

			p1 = p2;
			if (edgeId < 2)
			{
				p2 = pts[edgeId + 1];
			}
		}

  } // each cell

  cellIds->Delete();
	cellsVisited->Delete();
	cellsPoints->Delete();
	edgeTable->Delete();

}

void vtkSQ3SubdivisionFilter::GenerateSubdivisionCells (vtkPolyData *inputDS, 
																												vtkIntArray *edgeData, 
																												vtkCellArray *outputPolys, 
																												vtkCellData *outputCD)
{
  vtkIdType numCells = inputDS->GetNumberOfCells();
  vtkIdType cellId, newId;
  int id;
  vtkIdType npts;
  vtkIdType *pts;
  float edgePts[3];
  vtkIdType newCellPts[3];
  vtkCellData *inputCD = inputDS->GetCellData();

  // Now create new cells from existing points and generated edge points
  for (cellId=0; cellId < numCells; cellId++)
  {
    if ( inputDS->GetCellType(cellId) != VTK_TRIANGLE )
    {
      continue;
    }
    // get the original point ids and the ids stored as cell data
    inputDS->GetCellPoints(cellId, npts, pts);
    edgeData->GetTuple(cellId, edgePts);
    id = 0;
    newCellPts[id++] = pts[0];
    newCellPts[id++] = (int) edgePts[0];
    newCellPts[id++] = pts[2];
    newId = outputPolys->InsertNextCell (3, newCellPts);
    outputCD->CopyData (inputCD, cellId, newId);

    id = 0;
    newCellPts[id++] = pts[0];
    newCellPts[id++] = pts[1];
    newCellPts[id++] = (int) edgePts[0];
    newId = outputPolys->InsertNextCell (3, newCellPts);
    outputCD->CopyData (inputCD, cellId, newId);

    id = 0;
    newCellPts[id++] = pts[1];
    newCellPts[id++] = pts[2];
    newCellPts[id++] = (int) edgePts[0];
    newId = outputPolys->InsertNextCell (3, newCellPts);
    outputCD->CopyData (inputCD, cellId, newId);
	}
}

vtkIdType vtkSQ3SubdivisionFilter::InterpolatePosition (
        vtkPoints *inputPts, vtkPoints *outputPts,
        vtkIdList *stencil, float *weights)
{
  float *xx, x[3];
  int i, j;

  for (j = 0; j < 3; j++)
    {
    x[j] = 0.0;
    }

  for (i = 0; i < stencil->GetNumberOfIds(); i++)
    {
    xx = inputPts->GetPoint(stencil->GetId(i));
    for (j = 0; j < 3; j++)
      {
      x[j] += xx[j] * weights[i];
      }
    }
  return outputPts->InsertNextPoint (x);

}

int vtkSQ3SubdivisionFilter::FindEdge (vtkPolyData *mesh,
                                                 vtkIdType cellId,
                                                 vtkIdType p1, vtkIdType p2,
                                                 vtkIntArray *edgeData,
                                                 vtkIdList *cellIds)
{
  int edgeId = 0;
  int currentCellId = 0;
  int i;
  int numEdges;
  vtkIdType tp1, tp2;
  vtkCell *cell;

  // get all the cells that use the edge (except for cellId)
  mesh->GetCellEdgeNeighbors (cellId, p1, p2, cellIds);

  // find the edge that has the point we are looking for
  for ( i=0; i < cellIds->GetNumberOfIds(); i++)
    {
    currentCellId = cellIds->GetId(i);
    cell = mesh->GetCell(currentCellId);
    numEdges = cell->GetNumberOfEdges();
    tp1 = cell->GetPointId(2);
    tp2 = cell->GetPointId(0);
    for (edgeId=0; edgeId < numEdges; edgeId++)
      {
      if ( (tp1 == p1 && tp2 == p2) ||
           (tp2 == p1 && tp1 == p2))
        {
        break;
        }
      tp1 = tp2;
      tp2 = cell->GetPointId(edgeId + 1);
      }
    }
    // found the edge, return the stored value
    return (int) edgeData->GetComponent(currentCellId,edgeId);
}

void vtkSQ3SubdivisionFilter::GenerateEvenStencil (vtkIdType p1,
                                                    vtkPolyData *polys,
                                                    vtkIdList *stencilIds,
                                                    float *weights)
{
  vtkIdList *cellIds = vtkIdList::New();
  vtkIdList *ptIds = vtkIdList::New();
  vtkCell *cell;

  int i, j;
  int numCellsInLoop;
  int startCell, nextCell;
  vtkIdType p, p2;
  vtkIdType bp1, bp2;
  int K;
  float beta, cosSQ;

  // Get the cells that use this point
  polys->GetPointCells (p1, cellIds);
  numCellsInLoop = cellIds->GetNumberOfIds();
  if (numCellsInLoop < 1)
      {
      vtkWarningMacro("numCellsInLoop < 1: " << numCellsInLoop);
      stencilIds->Reset();
      return;
      }
  // Find an edge to start with that contains p1
  polys->GetCellPoints (cellIds->GetId(0), ptIds);
  p2 = ptIds->GetId(0);
  i = 1;
  while (p1 == p2)
    {
    p2 = ptIds->GetId(i++);
    }
  polys->GetCellEdgeNeighbors (-1, p1, p2, cellIds);

  nextCell = cellIds->GetId(0);
  bp2 = -1;
  bp1 = p2;
  if (cellIds->GetNumberOfIds() == 1)
    {
    startCell = -1;
    }
  else
    {
    startCell = cellIds->GetId(1);
    }

  stencilIds->Reset();
  stencilIds->InsertNextId(p2);

  // walk around the loop counter-clockwise and get cells
  for (j = 0; j < numCellsInLoop; j++)
    {
    cell = polys->GetCell(nextCell);
    p = -1;
    for (i = 0; i < 3; i++)
      {
      if ((p = cell->GetPointId(i)) != p1 && cell->GetPointId(i) != p2)
        {
        break;
        }
      }
    p2 = p;
    stencilIds->InsertNextId (p2);
    polys->GetCellEdgeNeighbors (nextCell, p1, p2, cellIds);
    if (cellIds->GetNumberOfIds() != 1)
      {
      bp2 = p2;
      j++;
      break;
      }
    nextCell = cellIds->GetId(0);
    }

  // now walk around the other way. this will only happen if there
  // is a boundary cell left that we have not visited
  nextCell = startCell;
  p2 = bp1;
  for (; j < numCellsInLoop && startCell != -1; j++)
    {
    cell = polys->GetCell(nextCell);
    p = -1;
    for (i = 0; i < 3; i++)
      {
      if ((p = cell->GetPointId(i)) != p1 && cell->GetPointId(i) != p2)
        {
        break;
        }
      }
    p2 = p;
    stencilIds->InsertNextId (p2);
    polys->GetCellEdgeNeighbors (nextCell, p1, p2, cellIds);
    if (cellIds->GetNumberOfIds() != 1)
      {
      bp1 = p2;
      break;
      }
    nextCell = cellIds->GetId(0);
    }

  if (bp2 != -1) // boundary edge
    {
    stencilIds->SetNumberOfIds(3);
    stencilIds->SetId(0,bp2);
    stencilIds->SetId(1,bp1);
    stencilIds->SetId(2,p1);
    weights[0] = .125;
    weights[1] = .125;
    weights[2] = .75;
    }
  else
    {
    K = stencilIds->GetNumberOfIds();
   // Remove last id. It's a duplicate of the first
    K--;
    if (K > 3)
      {
      // Generate weights
#define VTK_PI 3.14159265358979
      cosSQ = .375 + .25 * cos (2.0 * VTK_PI / (float) K);
      cosSQ = cosSQ * cosSQ;
      beta = (.625 -  cosSQ) / (float) K;
      }
    else
      {
      beta = 3.0 / 16.0;
      }
    for (j = 0; j < K; j++)
      {
      weights[j] = beta;
      }
    weights[K] = 1.0 - K * beta;
    stencilIds->SetId (K,p1);
    }
  cellIds->Delete();
  ptIds->Delete();
}

void vtkSQ3SubdivisionFilter::PrintSelf(ostream& os, vtkIndent indent)
{
  this->Superclass::PrintSelf(os,indent);

  os << indent << "Number of subdivisions: " << this->NumberOfSubdivisions << endl;
}


