/*=========================================================================

  Program:   Visualization Toolkit
  Module:    $RCSfile: vtkSQ3SubdivisionFilter.h,v $
  Language:  C++
  Date:      $Date: 2002/09/03 12:52:23 $
  Version:   $Revision: 1.16 $

  Copyright (c) 1993-2002 Ken Martin, Will Schroeder, Bill Lorensen 
  All rights reserved.
  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even 
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/
// .NAME vtkSQ3SubdivisionFilter - generate a subdivision surface using an Interpolating Scheme
// .SECTION Description
// vtkSQ3SubdivisionFilter is an abstract class that defines
// the protocol for interpolating subdivision surface filters.

// .SECTION Thanks
// This work was supported by PHS Research Grant No. 1 P41 RR13218-01
// from the National Center for Research Resources.

// .SECTION See Also
// vtkLinearSubdivisionFilter vtkButterflySubdivisionFilter

#ifndef __vtkSQ3SubdivisionFilter_h
#define __vtkSQ3SubdivisionFilter_h

#include <vtkPolyDataToPolyDataFilter.h>

class vtkCellArray;
class vtkCellData;
class vtkIdList;
class vtkIntArray;
class vtkPointData;
class vtkPoints;
class vtkPolyData;

/**
 * \htmlonly
 * SQ3 subdivision implements the &radic;3-subdivision implemented by Leif Kobbelt at Max-Planck
 * Institut for Computer Graphics. The filter can handle only triangle meshes so if you have other
 * mashes use the vtkTriangleFilter to triangulate it.
 * \endhtmlonly
 */
class vtkSQ3SubdivisionFilter : public vtkPolyDataToPolyDataFilter
{
public:
  static vtkSQ3SubdivisionFilter *New();
  vtkTypeRevisionMacro(vtkSQ3SubdivisionFilter,vtkPolyDataToPolyDataFilter);
  void PrintSelf(ostream& os, vtkIndent indent);

  // Description:
  // Set/get the number of subdivisions.
  vtkSetMacro(NumberOfSubdivisions,int);
  vtkGetMacro(NumberOfSubdivisions,int);

protected:
	vtkSQ3SubdivisionFilter () {
		this->NumberOfSubdivisions = 1;

		//os = new ofstream("C:\\vtk.log", ios::app | ios::out);
		/*
		os = new ofstream("C:\\vtk.log");
		(*os) << "\n\n\n============================ S T A R T ===============================\n";
		*/
	};
	~vtkSQ3SubdivisionFilter () { 
		/*
		(*os) << "\n\n\n============================= E N D E ================================\n";
		os->flush(); 
		os->close(); 
		delete os; 
		*/
	};

  void Execute();
	/**
	 * Generates a new point in the output mash. It also ubdates the cellsVisited and cellsPoints list.
	 * The new point is added to outputPD and outputPts.
	 */
	bool generatePoint(vtkIdType cellId, vtkIdType &newId, 
										 vtkIdList *cellsVisited, vtkIdList *cellsPoints, 
										 vtkIdType *pts, vtkPolyData *inputDS, 
										 vtkPointData *outputPD, vtkPoints *outputPts);
	/**
	 * Subdivides each cell and generates new cells. The methode does this by first filtering the input
	 * and then generating new points. It uses the generatePoints method.
	 *
	 * @todo	While stepping through all neighbours we assume there is only one neighbour per edge.
	 */
  void GenerateSubdivisionPoints (vtkPolyData *inputDS, vtkPoints *outputPts, vtkPointData *outputPD, 
																	vtkCellArray *outputPolys, vtkCellData *outputCD);
  void GenerateSubdivisionCells (vtkPolyData *inputDS, vtkIntArray *edgeData, vtkCellArray *outputPolys, vtkCellData *outputCD);
	/**
	 * Generates a filter stencil for input points.
	 */
	void GenerateEvenStencil (vtkIdType p1, vtkPolyData *polys,
                            vtkIdList *stencilIds, float *weights);

	/**
	 * Find a edge in mesh using p1 and p2.
	 */
  int FindEdge (vtkPolyData *mesh, vtkIdType cellId, vtkIdType p1,
                vtkIdType p2, vtkIntArray *edgeData, vtkIdList *cellIds);
	/**
	 * Calculates the position of a new point using the stencil and weights.
	 */
  vtkIdType InterpolatePosition (vtkPoints *inputPts, vtkPoints *outputPts,
                                 vtkIdList *stencil, float *weights);
  int NumberOfSubdivisions;

private:
	//ofstream *os;

  vtkSQ3SubdivisionFilter(const vtkSQ3SubdivisionFilter&);  // Not implemented.
  void operator=(const vtkSQ3SubdivisionFilter&);  // Not implemented.
};

#endif


